from __future__ import annotations


class PermissionType(type):
    def __or__(self, other: BasePermission | type[BasePermission]) -> BasePermission:
        return Or(self, other)

    def __and__(self, other: BasePermission | type[BasePermission]) -> BasePermission:
        return And(self, other)


class BasePermission(metaclass=PermissionType):
    def check_permission(self):
        raise NotImplementedError

    def __or__(self, other: BasePermission | type[BasePermission]) -> BasePermission:
        return Or(self, other)

    def __and__(self, other: BasePermission | type[BasePermission]) -> BasePermission:
        return And(self, other)

    def __call__(self) -> BasePermission:
        return self


class Permission(BasePermission):
    def has_permission(self) -> bool:
        ...

    @property
    def error_message(self) -> str:
        ...

    def check_permission(self):
        if not self.has_permission():
            raise PermissionError(self.error_message)


class Or(BasePermission):
    def __init__(self, *args: tuple[BasePermission | type[BasePermission], ...]):
        self.permissions = args

    def check_permission(self) -> bool:
        errors = []
        for permission in self.permissions:
            try:
                return permission().check_permission()
            except PermissionError as e:
                errors.append(e)
                continue

        if errors:
            raise errors[0] from None


class And(BasePermission):
    def __init__(self, *args: tuple[BasePermission | type[BasePermission], ...]):
        self.permissions = args

    def check_permission(self):
        for permission in self.permissions:
            try:
                permission().check_permission()
            except PermissionError:
                raise



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from flask import Flask


class BaseMiddleware:
    def __init__(self, app: Flask | None = None):
        self.app = app
        if app:
            self.init_app(app)

    def init_app(self, app: Flask):
        self.app = app
        self.app.before_request(self.before_request)
        self.app.after_request(self.after_request)

    def before_request(self):
        pass

    def after_request(self, response):
        return response



from __future__ import annotations

from typing import TYPE_CHECKING

from flask import jsonify, make_response

from ._utils import method

if TYPE_CHECKING:
    from flask.views import MethodView

    _Base = MethodView
else:
    _Base = object


class JsonifyMixin(_Base):
    def dispatch_request(self, *args, **kwargs):
        """
        Jsonify the dict or list of items in the response
        """
        response = super().dispatch_request(*args, **kwargs)

        if response is None:
            response = {}

        is_tuple = isinstance(response, tuple)

        if is_tuple:
            if response[0] is None:
                response = ({}, response[1])

            if isinstance(response[0], dict) or isinstance(response[0], list):
                return make_response(jsonify(response[0]), response[1])

        if isinstance(response, dict) or isinstance(response, list):
            return jsonify(response)

        return response


class StatusCodeMixin(_Base):
    def dispatch_request(self, *args, **kwargs):
        """
        If the response has no status code, supplement it.
        If the response is None, 204 is returned, and an additional mixin is required
        to handle the None response.
        """
        response = super().dispatch_request(*args, **kwargs)
        is_tuple = isinstance(response, tuple)

        if is_tuple:
            return response

        # Infer the status code
        status = 201 if method() == "post" else 200
        response = (response, status if response is not None else 204)
        return response



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, Protocol

from ._utils import method

if TYPE_CHECKING:
    from flask.views import MethodView

    _Base = MethodView
else:
    _Base = object


class PermissionProtocol(Protocol):
    def __init__(self, *args, **kwargs):
        ...

    def check_permission(self):
        """Check that the permission is satisfied"""


class PermissionMixin(_Base):
    def _get_permissions(self) -> Iterable[type[PermissionProtocol]]:
        return getattr(self, "permissions", [])

    def get_write_permissions(self) -> Iterable[type[PermissionProtocol]]:
        return self._get_permissions()

    def get_read_permissions(self) -> Iterable[type[PermissionProtocol]]:
        return self._get_permissions()

    def get_post_permissions(self) -> Iterable[type[PermissionProtocol]]:
        return self.get_write_permissions()

    def get_put_permissions(self) -> Iterable[type[PermissionProtocol]]:
        return self.get_write_permissions()

    def get_patch_permissions(self) -> Iterable[type[PermissionProtocol]]:
        return self.get_write_permissions()

    def get_delete_permissions(self) -> Iterable[type[PermissionProtocol]]:
        return self.get_write_permissions()

    def get_get_permissions(self) -> Iterable[type[PermissionProtocol]]:
        return self.get_read_permissions()

    def get_permissions(self) -> Iterable[type[PermissionProtocol]]:
        if method_ := getattr(self, f"get_{method()}_permissions", None):
            return method_()
        return self._get_permissions()

    def dispatch_request(self, *args, **kwargs) -> Any:
        for permission in self.get_permissions():
            permission().check_permission()

        return super().dispatch_request(*args, **kwargs)



from __future__ import annotations

from flask import request
from flask.views import MethodView

from .view_mixins import (
    JsonifyMixin,
    PermissionMixin,
    SchemaMixin,
    ServiceMixin,
    StatusCodeMixin,
)


class _BaseView(
    JsonifyMixin,
    ServiceMixin,
    StatusCodeMixin,
    SchemaMixin,
    PermissionMixin,
    MethodView,
):
    pass


class ResourceView(_BaseView):
    pass


class ResourcesView(_BaseView):
    def get_response_schema_options(self) -> dict:
        return {"many": request.method.lower() == "get"}



from .misc_mixins import JsonifyMixin, StatusCodeMixin
from .permission_mixin import PermissionMixin
from .schema_mixin import SchemaMixin
from .service_mixin import ServiceMixin

__all__ = [
    "JsonifyMixin",
    "StatusCodeMixin",
    "PermissionMixin",
    "SchemaMixin",
    "ServiceMixin",
]



from __future__ import annotations

from typing import Any, Dict


class ServiceMixin:
    service_class = None

    def get_service_options(self, *args, **kwargs) -> Dict[Any, Any]:
        # Can be overridden
        return {}

    def get_service(self, *args, **kwargs) -> Any:
        # Can be overridden
        if not self.service_class:
            raise RuntimeError("Unable to get service with no service_class defined")

        service_class: type[Any] = self.service_class
        service_options = self.get_service_options(*args, **kwargs)
        service_options.update(**kwargs)
        return service_class(*args, **service_options)



from distutils.version import LooseVersion

from .middleware import BaseMiddleware
from .permissions import BasePermission, Permission
from .view_mixins.misc_mixins import JsonifyMixin, StatusCodeMixin
from .view_mixins.permission_mixin import PermissionMixin
from .view_mixins.schema_mixin import SchemaMixin
from .view_mixins.service_mixin import ServiceMixin
from .views import ResourcesView, ResourceView

__all__ = [
    "JsonifyMixin",
    "StatusCodeMixin",
    "PermissionMixin",
    "SchemaMixin",
    "ServiceMixin",
    "ResourceView",
    "ResourcesView",
    "BasePermission",
    "Permission",
    "BaseMiddleware",
]

__version__ = "0.0.6"
__version_info__ = tuple(LooseVersion(__version__).version)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from flask import request

from ._utils import method

if TYPE_CHECKING:
    from marshmallow import Schema


if TYPE_CHECKING:
    from flask.views import MethodView

    _Base = MethodView
else:
    _Base = object


class _FilterSchemaMixin:
    filter_schema = None

    def get_filter_schema_class(self) -> type[Schema]:
        # Can be overridden
        if not self.filter_schema:
            raise RuntimeError("No filter schema defined for the class")

        return self.filter_schema

    def get_filter_schema_context(self) -> dict:
        # Can be overridden
        return {}

    def get_filter_schema_options(self) -> dict:
        # Can be overridden
        return {}

    def get_filter_schema_instance(self) -> Schema:
        # Can be overridden
        class_ = self.get_filter_schema_class()
        return class_(
            context=self.get_filter_schema_context(), **self.get_filter_schema_options()
        )

    def get_filter_data(self) -> dict | Any:
        return self.get_filter_schema_instance().load(request.args.to_dict())


class _ResponseSchemaMixin(_Base):
    schema = None
    response_schema = None

    def get_response_schema_class(self, *args, **kwargs) -> type[Schema]:
        # Can be overridden
        if schema := self.response_schema:
            return schema

        if not self.schema:
            raise RuntimeError("No response schema defined in the class")

        return self.schema

    def get_response_schema_context(self) -> dict:
        # Can be overridden
        return {}

    def get_response_schema_options(self) -> dict:
        # Can be overridden
        return {}

    def get_response_schema_instance(self) -> Schema:
        # Can be overridden
        class_ = self.get_response_schema_class()
        return class_(
            context=self.get_response_schema_context(),
            **self.get_response_schema_options(),
        )

    @property
    def _many_response(self) -> bool:
        return self.get_response_schema_options().get("many", False)

    def dispatch_request(self, *args, **kwargs):
        """
        Convert the response to a dict or list of dicts using the response schema
        if there is a response object. Keep the status code if there is one, and
        return an empty dictionary if the given object is None
        """
        response = super().dispatch_request(*args, **kwargs)
        tuple_response = isinstance(response, tuple)

        if response is None or tuple_response and response[0] is None:
            return response

        schema = self.get_response_schema_instance()

        obj = response[0] if tuple_response else response
        should_be_many = self._many_response
        should_be_single = not should_be_many

        if not isinstance(response, dict):
            if should_be_single and isinstance(response, list):
                raise RuntimeError(
                    "View returned list, but expected an individual item"
                )
            if should_be_many and not isinstance(response, list):
                raise RuntimeError("View returned non-list, but expected list")

            obj = schema.dump(response)

        return (obj, response[1]) if tuple_response else obj


class _RequestSchemaMixin:
    request_schema = None
    schema = None

    def get_request_schema_context(self) -> dict:
        # Can be overridden
        return {}

    def get_request_schema_options(self) -> dict:
        # Can be overridden
        return {}

    def get_request_schema_class(self, *args, **kwargs) -> type[Schema]:
        # Can be overridden
        if schema := self.request_schema:
            return schema

        if not self.schema:
            raise RuntimeError("No request schema defined in the class")

        return self.schema

    def get_request_schema_instance(self) -> Schema:
        class_ = self.get_request_schema_class()
        return class_(
            context=self.get_request_schema_context(),
            **self.get_request_schema_options(),
        )

    def get_patch_schema_class(self) -> type[Schema]:
        # Can be overridden
        return self.get_request_schema_class()

    def get_patch_schema_context(self) -> dict:
        # Can be overridden
        return self.get_request_schema_context()

    def get_patch_schema_options(self) -> dict:
        # Can be overridden
        return self.get_request_schema_options()

    def get_patch_schema_instance(self) -> Schema:
        # Can be overridden
        class_ = self.get_patch_schema_class()
        return class_(
            context=self.get_patch_schema_context(), **self.get_patch_schema_options()
        )

    def get_post_schema_class(self) -> type[Schema]:
        # Can be overridden
        return self.get_request_schema_class()

    def get_post_schema_context(self) -> dict:
        # Can be overridden
        return self.get_request_schema_context()

    def get_post_schema_options(self) -> dict:
        # Can be overridden
        return self.get_request_schema_options()

    def get_post_schema_instance(self) -> Schema:
        # Can be overridden
        class_ = self.get_post_schema_class()
        return class_(
            context=self.get_post_schema_context(), **self.get_post_schema_options()
        )

    def get_put_schema_class(self) -> type[Schema]:
        # Can be overridden
        return self.get_request_schema_class()

    def get_put_schema_context(self) -> dict:
        # Can be overridden
        return self.get_request_schema_context()

    def get_put_schema_options(self) -> dict:
        # Can be overridden
        return self.get_request_schema_options()

    def get_put_schema_instance(self) -> Schema:
        # Can be overridden
        class_ = self.get_put_schema_class()
        return class_(
            context=self.get_put_schema_context(), **self.get_put_schema_options()
        )

    def get_get_schema_class(self) -> type[Schema]:
        # Can be overridden
        return self.get_request_schema_class()

    def get_get_schema_context(self) -> dict:
        # Can be overridden
        return self.get_request_schema_context()

    def get_get_schema_options(self) -> dict:
        # Can be overridden
        return self.get_request_schema_options()

    def get_get_schema_instance(self) -> Schema:
        # Can be overridden
        class_ = self.get_get_schema_class()
        return class_(
            context=self.get_get_schema_context(), **self.get_get_schema_options()
        )

    def _get_request_schema_instance(self) -> Schema:
        if method_ := getattr(self, f"get_{method()}_schema_instance", None):
            return method_()
        return self.get_request_schema_instance()

    def get_validated_data(self) -> dict | Any:
        data = request.get_json(force=True)
        if data is not None:
            return self._get_request_schema_instance().load(data)
        return {}


class SchemaMixin(_RequestSchemaMixin, _ResponseSchemaMixin, _FilterSchemaMixin):
    pass



import re

from setuptools import find_packages, setup


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content


EXTRAS_REQUIRE = {}
# EXTRAS_REQUIRE["tests"] = read("requirements/test.requirements.txt").splitlines()
# EXTRAS_REQUIRE["dev"] = read("requirements/dev.requirements.txt").splitlines()
REQUIRES = read("requirements/requirements.txt").splitlines()


def find_version(fname):
    """Attempts to find the version number in the file names fname.
    Raises RuntimeError if not found.
    """
    version = ""
    with open(fname) as fp:
        reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
        for line in fp:
            m = reg.match(line)
            if m:
                version = m.group(1)
                break
    if not version:
        raise RuntimeError("Cannot find version information")
    return version


setup(
    name="flask-mixins",
    version=find_version("src/flask_mixins/__init__.py"),
    description="Flask and marshmallow cleaner code",
    author="Tomas Sheers",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author_email="t.sheers@outlook.com",
    url="https://github.com/reritom/flask-mixins",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    license="MIT",
    zip_safe=False,
    keywords="flask-mixins",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    test_suite="tests",
    project_urls={
        "Issues": "https://github.com/reritom/flask-mixins/issues",
    },
)
